/*****************************************************************************/

/*
 *
 *   Copyright (c) 2002, Smart Link Ltd.
 *   All rights reserved.
 *
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *       1. Redistributions of source code must retain the above copyright
 *          notice, this list of conditions and the following disclaimer.
 *       2. Redistributions in binary form must reproduce the above
 *          copyright notice, this list of conditions and the following
 *          disclaimer in the documentation and/or other materials provided
 *          with the distribution.
 *       3. Neither the name of the Smart Link Ltd. nor the names of its
 *          contributors may be used to endorse or promote products derived
 *          from this software without specific prior written permission.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *   OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *   OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 *
 *	sysdep_amr.c  --  pci driver sysdep routines.
 *
 *	Author: Seva (seva@smlink.com)
 *
 *
 */

/*****************************************************************************/


#include <linux/version.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <asm/io.h>

/* memory allocations */
void *sysdep_malloc (unsigned int size)
{
        void *mem;
        mem = kmalloc(size, in_interrupt() ? GFP_ATOMIC : GFP_KERNEL);
        if(!mem)
                printk(KERN_ERR "slmdm: cannot alloc %db mem\n", size);
        return mem;
}

void sysdep_free (void *mem)
{
	kfree(mem);
}

/* strings */
void *sysdep_memset(void *d,int c,size_t l)
{
	return memset(d,c,l);
}

void *sysdep_memcpy(void *d,const void *s,size_t l)
{
	return memcpy(d,s,l);
}

size_t sysdep_strlen(const char *s)
{
	return strlen(s);
}

char *sysdep_strcpy(char *d,const char *s)
{
	return strcpy(d,s);
}

char *sysdep_strncpy(char *d,const char *s, size_t l)
{
	return strncpy(d,s,l);
}

int sysdep_strcmp(const char *s1,const char *s2)
{
	return strcmp(s1, s2);
}

int sysdep_sprintf(char *buf, const char *fmt, ...)
{
	va_list args;
	int i;
	va_start(args, fmt);
	i=vsprintf(buf,fmt,args);
	va_end(args);
	return i;
}

/* getting times */
unsigned long sysdep_get_utime(void)
{
    struct timeval t;
    do_gettimeofday (&t);
    return t.tv_sec * 1000000 + t.tv_usec ;
}

/* cli/sti handling */
void sysdep_save_flags   (unsigned long *flags) {save_flags(*flags); }
void sysdep_restore_flags(unsigned long *flags) {restore_flags(*flags);}
void sysdep_cli(void) {cli();}


/* dmesg print */
void sysdep_dmesg_print(const char *s)
{
	printk("%s\n",s);
}

/* page allocations */
unsigned long sysdep_get_free_pages(unsigned long order)
{
	return __get_free_pages(GFP_KERNEL, order);
}

void sysdep_free_pages(unsigned long addr, unsigned long order)
{
	free_pages(addr, order);
}

unsigned long sysdep_virt_to_phys(volatile void * address)
{
	return virt_to_phys (address);
}

/* delay */
void sysdep_udelay(unsigned long usecs)
{
	udelay(usecs);
}

void sysdep_mdelay(unsigned long msecs)
{
	mdelay(msecs);
}

/* from linux/pci.h */
int sysdep_pci_read_config_byte (void *dev, unsigned char where, unsigned char *val)
{	
        return pci_read_config_byte (dev, where, val);
}

int sysdep_pci_read_config_word (void *dev, unsigned char where, unsigned short *val)
{
        return pci_read_config_word (dev, where, val);
}

int sysdep_pci_read_config_dword (void *dev, unsigned char where, unsigned int *val)
{
        return pci_read_config_dword (dev, where, val);
}

int sysdep_pci_write_config_byte (void *dev, unsigned char where, unsigned char val)
{
        return pci_write_config_byte (dev, where, val);
}

int sysdep_pci_write_config_word (void *dev, unsigned char where, unsigned short val)
{
        return pci_write_config_word (dev, where, val);
}

int sysdep_pci_write_config_dword (void *dev, unsigned char where, unsigned int val)
{
	return pci_write_config_dword (dev, where, val);
}

/* io */
unsigned char sysdep_inb(unsigned short port)
{
	return inb(port);
}
void sysdep_outb(unsigned char data,unsigned short port)
{
	outb(data,port);
}
unsigned short sysdep_inw(unsigned short port) 
{
	return inw(port);
}

void sysdep_outw(unsigned short data,unsigned short port)
{
	outw(data,port);
}
unsigned long sysdep_inl(unsigned short port) 
{
	return inl(port);
}
void sysdep_outl(unsigned long data,unsigned short port)
{
	outl(data,port);
}


